// CExec.c
// Contains A, B & C phases for a simple array-based 3 phase executive
#include "CExec.h"
#include "ExecUtil.h"

//#include <values.h>
#include <conio.h>
#include <stdio.h>
#include <dos.h>
#include <stdlib.h>


struct EntDetails Details[MaxEnt];			// Array of entity information

void (*CList[MaxC])(void);					 	// Array of C activities

int NumCurrEnts;                     		//	Used in APhase & BPhase
int Clock;                    				// Current Simulation time
int PrevClock;                				// Previous simulation time
int RunDuration;                 			// Length of run
int NumEnts;                     			// Number of entities
int NumCs;                       			// Number of C Activities
int CurrEnt;								// Entity causing current B
int CurrEntArray[MaxEnt]; 					// Entities due now
enum Boolean CStarted;                    // Used in CPhase for repeat scans
enum Boolean DeBug;                       // Is this a debug run?


void APhase(void);
void BPhase(void);
void CPhase(void);
void CheckInterrupt(void);

extern void UpDateScreen(void);				// Needs to be declared in Model.c
extern enum Boolean TraceOn;
extern FilePtr TraceFile;

void APhase(void)
// Time scan by examining the Details array of records.
// Ignores the first (0) element for consistency with entity numbers.
// Checks all entities for which Avail is false.
// For these, finds those with the smallest timecells
// and puts the entity number in the CurrEntArray.
{
	int Entity, Minm;

	NumCurrEnts = 0;
	Minm = MAXINT;
	for (Entity = 1; Entity <= NumEnts; Entity++)
		if (!Details[Entity].Avail)
			if (Details[Entity].TimeCell <= Minm)
			{
				if (Details[Entity].TimeCell < Minm)
					NumCurrEnts = 1;
					else
						NumCurrEnts++;
				Minm = Details[Entity].TimeCell;
				CurrEntArray[NumCurrEnts] = Entity;
			}
	if (Minm == MAXINT)
		Error("Error in A phase, Minm still = MaxInt");
	else if (Minm < 0)
		Error("Seems to be a mistake in A phase, Minm -ve");
	PrevClock = Clock;
	Clock = Minm;
}

void BPhase(void)
// Works through CurrEntArray. Ignores the first (0) element.
// For each entity number stored in CurrEntArray, executes the B activity
// indicated for that entity.
// Executed after A phase and before C phase.
{
	int Loop;

	if (TraceOn)
	{
		fprintf(TraceFile, "\n");
		fprintf(TraceFile, "%s, %5d ", "Time now ..", Clock);
		fprintf(TraceFile, "\n");
	}
	ShowEntDetails();

	for (Loop = 1; Loop <= NumCurrEnts; Loop++)
	{
		CurrEnt = CurrEntArray[Loop];
		Details[CurrEnt].Avail = True;					  	// Put entity avail
		(*Details[CurrEnt].NextB)();                   	// Do this B Activity
	}
}

void CPhase(void)
// Try each C activity in turn
// Ignores the first (0) element for compatibility.
{
	int Loop;
	CStarted = True;
	while (CStarted)
	{
		CStarted = False;
		for (Loop = 1; Loop <= NumCs; Loop++)
			(*CList[Loop])();
	}
}

void CheckInterrupt(void)
// Checks whether user has hit the keyboard. If so, pauses.
// Then asks for continuation or termination
{
	char Ch;
	if (_kbhit())
	{
		Ch = _getch();               // Empty the keyboard buffer
		_cprintf("\n\n");
		_cprintf("Press <ESC> to halt, any other key to continue");
		Ch = _getch();
		if (Ch == 27)
		{
			Finalisation();
			exit(99);
		}
	}
}

void Simulate(void)
// Main simulation loop
{
	Initialisation();
	while (Clock <= RunDuration)
	{
		APhase();
		BPhase();
		CPhase();
		UpDateScreen();
		CheckInterrupt();
	}
	Finalisation();
}

