// Executil.c
// Utilities for CExec

#include "GenLib.h"
#include "CExec.h"
#include "ExecUtil.h"

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <math.h>

extern int NumEnts, Clock, PrevClock, NumCs;
extern struct EntDetails Details[];
extern void (*CList[])(void);

FilePtr TraceFile;

enum Boolean TraceOn;

void Display(String Blurb1, int Int1, String Blurb2, int Int2)
// Writes to TraceFile if TraceOn
{
	if (TraceOn)
		fprintf(TraceFile, "%s %d %s %d \n", Blurb1, Int1, Blurb2, Int2);
}

void ShowEntDetails()
// Writes current state of entity details to the TraceFile
{
	int Entity;
	if (TraceOn)
		for (Entity = 1; Entity <= NumEnts; Entity++)
		{
			fprintf(TraceFile, "%d %s %s %d", Entity, ":", Details[Entity].Name, Entity[Details].TimeCell);
				if (Entity[Details].Avail)
					fprintf(TraceFile, "  Avail \n");
				else
					fprintf(TraceFile, "  Engaged \n");
		}
}

void Schedule(int Entity, void (*BAct)(), float RTime)
// Commits specified entity to Activity after RTime
{
	  Details[Entity].NextB = BAct;
	  Details[Entity].TimeCell = Clock + (int)RTime;
	  Details[Entity].Avail = False;
	  Details[Entity].Util += (int)RTime;
}

void Release(int Entity)
// Sets specified entity as Avail
{
	Details[Entity].Avail = True;
}

void InitEntFree(String Nom)
// Sets the next entity as initially free
// Keeps track of how many entities have been created so far
{

	if (NumEnts >= MaxEnt)
		Error("Tried to create too many entities");

	++NumEnts;

	Details[NumEnts].Name = Nom;
	Details[NumEnts].TimeCell = 0;
	Details[NumEnts].Util = 0;
	Release(NumEnts);
}

void InitEnt(void (*BAct)(), float RTime, String Nom)
// Sets up intitial state of the next entity
{
	InitEntFree(Nom);
	Schedule(NumEnts, BAct, RTime);
}

void AddC(void (*CAct)())
// Adds a C to the CArray
{
	if (NumCs >= MaxC)
		Error("Too many Cs in list: AddC, ExecUtil.c");
	NumCs++;
	CList[NumCs] = CAct;
}

void ZeroLimits()
//Sets Clock, NumEnts, NumCs to zero
{
	Clock = 0;                    		// Set simulation clock to zero
	PrevClock = 0;                     	// Set previous clock to zero
	NumEnts = 0;                       	// No entities created yet
	NumCs = 0;                         	// No Cs added yet
}

int GetObsInt()
// Gets observation interval as 10, 20 or 40
{
	int Value;
	Value = 0;
	while (!((Value == 10)||(Value == 20)||(Value == 40)))
	{
		_cprintf("Observation interval (10, 20 or 40): ");
		_cscanf("%i", &Value);
	}
	_cprintf("\n");
	return Value;
}

void InterruptMessage(void)
// Displays a message on-screen which invites a keypress in interrupt
{
	_cprintf(" ** PRESS ANY KEY TO INTERRUPT ** ");
}


float RandNum(void)
// Generates a floaing point pseudo random variate
// Unstreamed, as uses stdlib Rand() function
{
	float X, Y, Z;
	X = (float)rand();
	Y = (float)RAND_MAX;
	Z = X/Y;
	return Z;
}

float NegExp(int Mean)
// Returns a floating point sample from a neg exp distribution
// Takes an integer mean, uses unsatisfactory RandNum function
{
	return -1.0*(float)Mean*log(RandNum());
}

void EventTraceQuery()
// Asks user whether to produce an event trace file
{
	String FileName[255];
	char Ans;
	TraceOn = False;
	Ans = GetYNAns("Set up event trace file?");
	if ((Ans == 'Y')||(Ans == 'y'))
	{
		_cprintf("\n");
		_cprintf("File name? ");
		_cscanf("%s", FileName);
		if ((TraceFile = fopen(FileName, "w")) == NULL)
			Error("Unable to open output file");
		TraceOn = True;
	}
}
