// Model.c
// Straightforward Bs & Cs for C version of simple harassed booking clerk
// To be linked to CExec

#include "GenLib.h"
#include "CExec.h"
#include "ExecUtil.h"
#include "Model.h"

#include <stdlib.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>

extern enum Boolean CStarted, TraceOn;
extern struct EntDetails Details[];
extern int Clock, CurrEnt, NumEnts, RunDuration, DisplayDelay;
extern FilePtr TraceFile;


// ----------------------------------------------------------------------------
// -------------------------- Simulation variables ----------------------------
// ----------------------------------------------------------------------------

const int PersEnq = 1; 						// Entity number for Personal enquirere
const int PhoneEnq = 2;						// Entity number for phone callers
const int Observer = 3;						// Entity number for observer
const int FirstClerk = 4;					// Entity number for first clerk

int Obs;											// Count of observations

int PersIn;										// Count of personal arrivals
int PhoneIn;									// Count of phone arrivals
int PersOut;									// Count of personal departures
int PhoneOut;									// Count of phone departures

int PersQ;										// Sixe of personal queue
int PhoneQ;										// Size of phone queue

int PersArrTime;								// Mean personal IAT
int PhoneArrTime;								// Mean phone IAT
int PersServeTime;							// Mean personal service time
int PhoneServeTime;							// Mean phone service time

int ObsInterval;								// Interval between observations

Observations PersRec;						// Array of PersQ observations
Observations PhoneRec;						// Array of PhoneQ observations

int NumClerks;                   		//	Number of Clerks
int LastClerk;       						// ID of last clerk

// ----------------------------------------------------------------------------
// -------------------------------- Run-time stuff ----------------------------
// ----------------------------------------------------------------------------
void ScreenHeader()
// Puts header on screen for runtime output
{
	system("cls");
	_cprintf("Harassed booking clerk simulation - C version - press <ESC> to interrupt\n");
	_cprintf("------------------------------------------------------------------------\n\n");

	_cprintf("Number of clerks:  %d", NumClerks, "\n");
	_cprintf("\n\n");

	_cprintf("CLOCK   PERSONAL ENQUIRERS     PHONE CALLERS\n");
	_cprintf("          in    out    Q         in    out    Q\n");
}

void UpDateScreen()
// Displays run-time output
{
	static int Count = 0;
	if ((Clock % ObsInterval) == 0)
	{
		if ((Count % 15) == 0)
			ScreenHeader();
		Count++;

		_cprintf("%3d", Clock);
		_cprintf("%9d", PersIn);
		_cprintf("%7d", PersOut);
		_cprintf("%5d", PersQ);
		_cprintf("%11d", PhoneIn);
		_cprintf("%7d", PhoneOut);
		_cprintf("%5d", PhoneQ);

		_cprintf("\n");

		if (((Count % 15) == 0)||(Clock+ObsInterval)>RunDuration)
			AnyKey();
	}
}


// ----------------------------------------------------------------------------
// -------------------------------- B Activities ------------------------------
// ----------------------------------------------------------------------------
void Arrive(void)
// Arrival of personal enquirer
{
	PersIn++;    								// Increment number of arrivals
	PersQ++;        							// Increment queue length
	Display("Personal enquirer No ", PersIn, " arrives, queue = ", PersQ);
	Schedule(CurrEnt, Arrive, NegExp(PersArrTime)); // Schedule next persenq
}

void EndServe(void)
// End of personal service. CurrEnt is clerk who is released
{
	int ClerkNum;
	ClerkNum = CurrEnt - FirstClerk +1;
	PersOut++;    								// Increment number completed
	Display("End of personal service:", PersOut, " freed Clerk:", ClerkNum);
}

void Call(void)
// Arrival of phone call
{
	PhoneIn++;  								// Increment number of calls
	PhoneQ++;      							// Increment queue length}
	Display("Phone enquirer No. ", PhoneIn, " arrives, queue = ", PhoneQ);
	Schedule(CurrEnt, Call, NegExp(PhoneArrTime));  // Schedule next call
}

void EndTalk(void)
// End of phone conversation.
{
	int ClerkNum;
	ClerkNum = CurrEnt - FirstClerk + 1;
	PhoneOut++;    							// Increment number completed
	Display("End of phone conversation:", PhoneOut, " freed Clerk:", ClerkNum);
}

void Observe(void)
// Collect time series of queue lengths
{
	PersRec[Obs] = PersQ;
	PhoneRec[Obs] = PhoneQ;
	Obs++;
	Schedule(CurrEnt, Observe, (float)ObsInterval);    // Schedule next observation}
	Display("Recording, Personal queue = ", PersQ, ", phone queue = ",PhoneQ);
}

// ----------------------------------------------------------------------------
// -------------------------------- C Activities ------------------------------
// ----------------------------------------------------------------------------
void BeginPersService(void)
// Serve first customer in the queue of personal enquirers
{
	int ThisClerk;
	ThisClerk = FirstClerk-1;
	while ((PersQ > 0) && (ThisClerk < LastClerk))
	{
		ThisClerk++;
		if (Details[ThisClerk].Avail)
		{
			CStarted = True;
			PersQ--;
			Schedule(ThisClerk, EndServe, NegExp(PersServeTime));
			Display("Start Personal service with Clerk: ", ThisClerk, ", PersQ = ", PersQ);
		}
	}
}

void BeginPhoneConv(void)
// Answer call of first caller in the call queue
{
	int ThisClerk;
	ThisClerk = FirstClerk - 1;
	while ((PhoneQ > 0) && (ThisClerk < LastClerk))
	{
		ThisClerk++;
		if (Details[ThisClerk].Avail)
		{
			CStarted = True;
			PhoneQ--;
			Schedule(ThisClerk, EndTalk, NegExp(PhoneServeTime));
			Display("Start Phone Service with Clerk: ", ThisClerk, ", PhoneQ = ", PhoneQ);
		}
	}
}

// ----------------------------------------------------------------------------
// --------------------- Initialisation & Finalisation ------------------------
// ----------------------------------------------------------------------------

void GetRunPars()
// Asks for the parameters of the simulation run.
// Called by initialisation
{
	system("cls");
	_cprintf("Harassed booking clerk simulation - C version\n");
	_cprintf("---------------------------------------------\n\n");

	NumClerks = GetInt("How many clerks?", 1, 5);
	LastClerk = 3+NumClerks;    	  	// ID of LastClerk in Details array}
	NumEnts = LastClerk;        		// No more entities than this
	RunDuration = GetInt("Simulation duration?", 20, 1440);

	_cprintf("\nInput mean inter-arrival times\n");
	PersArrTime = GetInt("Personal enquirers", 1, 50);
	PhoneArrTime = GetInt("Phone callers", 1, 50);

	_cprintf("\nInput mean service times\n");
	PersServeTime = GetInt("Personal enquirers", 1, 50);
	PhoneServeTime = GetInt("Phone callers", 1, 50);

	_cprintf("\n");
	ObsInterval = GetObsInt();

	EventTraceQuery();
}

void InitCounters()
// Initialises the counters used by the model
{
	PersIn = 0;
	PhoneIn = 0;
	PersOut = 0;
	PhoneOut = 0;
	PersQ = 0;
	PhoneQ = 0;
}

void InitObservations()
// Sets up time series for data collection
{
	int Loop;
	for (Loop = 0; Loop < MaxObs; Loop++)
	{
		PersRec[Loop] = 0;
		PhoneRec[Loop] = 0;
	}
	Obs = 0;
}

void Initialisation()
// Sets up simulation ready to run
{
	int Clerk;

	GetRunPars();                           		// Get run parameters
	ZeroLimits();                           		// Set parameters to zero
	InitCounters();                         		// Set counters to zero
	InitObservations();                     		// Set up time series
//	Set up first arrivals & first observation of queue lengths}
	InitEnt(Arrive, NegExp(PersArrTime), "Personal  ");	// PersEnq
	InitEnt(Call, NegExp(PhoneArrTime), "Phone     ");		// PhoneEnq
	InitEnt(Observe, (float)ObsInterval, "Observer  ");				// Observer

	for (Clerk = FirstClerk; Clerk <= LastClerk; Clerk++)
		InitEntFree("Clerk");                           	// Clerks to idle

	AddC(BeginPersService);
	AddC(BeginPhoneConv);
}

void Finalisation()
// Final results of simulation
{
	int Clerk;
	int Counter;
	int Time;
	float Utilisation;

	if (TraceOn)
		fclose(TraceFile);

	system("cls");
	_cprintf("Harassed booking clerk simulation - C version\n");
	_cprintf("---------------------------------------------\n\n");


	_cprintf("The input mean times were ..\n");
	_cprintf("Personal customers:\n");
	_cprintf("   inter-arrival time: %5d", PersArrTime);
	_cprintf("   service time: %5d", PersServeTime);
	_cprintf("\n");
	_cprintf("Phone callers:\n");
	_cprintf("   inter-arrival time: %5d", PhoneArrTime, "\n");
	_cprintf("   service time: %5d", PhoneServeTime, "\n");

	_cprintf("\nClerk percent utilisations were as follows ..\n");
	Clerk = FirstClerk - 1;
	for (Counter = 1; Counter <= NumClerks; Counter++)
	{
		Clerk++;
		Utilisation = 100.0*((float)Details[Clerk].Util)/((float)Clock);
		_cprintf("%5d %8.0f", Counter, Utilisation);
		_cprintf("\n");
	}
	AnyKey();

	_cprintf("\n");

	printf("Time    Pers Queue     Phone Queue\n");
	printf("----    ----------     -----------\n");
	for (Counter = 1; Counter < Obs; Counter++)
	{
		Time = Counter*ObsInterval;
		if ((Counter % 20) == 0)        //Avoid losing table before seeing it
		{
			AnyKey();
			system("cls");
			printf("Time    Pers Queue     Phone Queue\n");
			printf("----    ----------     -----------\n");
		}
		printf("%3d %11d %15d \n", Time, PersRec[Counter], PhoneRec[Counter]);
	}
	printf("\n");
	printf("%s %d", "Simulation over after ", Clock);
	printf("\n");
}

void main()
// Main loop for this model
{
	Simulate();
}
